from sqlalchemy import create_engine, engine

from aqueduct_executor.operators.connectors.tabular import config, relational


class PostgresConnector(relational.RelationalConnector):
    def __init__(self, config: config.PostgresConfig):
        conn_engine = _create_engine(config)
        super().__init__(conn_engine)


def _create_engine(config: config.PostgresConfig) -> engine.Engine:
    # Postgres Dialect:
    # https://docs.sqlalchemy.org/en/14/dialects/postgresql.html#module-sqlalchemy.dialects.postgresql.psycopg2
    url = "postgresql://{username}:{password}@{host}:{port}/{database}".format(
        username=config.username,
        password=config.password,
        host=config.host,
        port=config.port,
        database=config.database,
    )
    return create_engine(url)



"""
CONFIG FILE FOR TABULAR CONNECTOR INTEGRATION TESTS
- To skip a particular test set the relevant `SKIP_` flag to False
- Set the `_CONF` dict for all connectors being tested.
- Commented out config fields are optional.
"""
from aqueduct_executor.operators.connectors.tabular import postgres
from aqueduct_executor.operators.connectors.tabular import snowflake
from aqueduct_executor.operators.connectors.tabular import mysql
from aqueduct_executor.operators.connectors.tabular import sql_server
from aqueduct_executor.operators.connectors.tabular import bigquery
from aqueduct_executor.operators.connectors.tabular import sqlite

"""FLAGS TO SKIP TESTS"""
SKIP_POSTGRES = True
SKIP_SNOWFLAKE = True
SKIP_MYSQL = True
SKIP_REDSHIFT = True
SKIP_MARIADB = True
SKIP_SQL_SERVER = True
SKIP_BIGQUERY = True
SKIP_SQLITE = True

"""POSTGRES CONFIG"""
POSTGRES_CONF = {
    postgres._CONFIG_USERNAME_KEY: "",
    postgres._CONFIG_PASSWORD_KEY: "",
    postgres._CONFIG_DATABASE_KEY: "",
    postgres._CONFIG_HOST_KEY: "",
    # postgres._CONFIG_PORT_KEY: "",
}

"""SNOWFLAKE CONFIG"""
SNOWFLAKE_CONF = {
    snowflake._CONFIG_USERNAME_KEY: "",
    snowflake._CONFIG_PASSWORD_KEY: "",
    snowflake._CONFIG_ACCOUNT_IDENTIFIER_KEY: "",
    snowflake._CONFIG_DATABASE_KEY: "",
    snowflake._CONFIG_WAREHOUSE_KEY: "",
    # snowflake._CONFIG_SCHEMA_KEY: "",
}

"""MYSQL CONFIG"""
MYSQL_CONF = {
    mysql._CONFIG_USERNAME_KEY: "",
    mysql._CONFIG_PASSWORD_KEY: "",
    mysql._CONFIG_DATABASE_KEY: "",
    mysql._CONFIG_HOST_KEY: "",
    mysql._CONFIG_PORT_KEY: "",
}

"""REDSHIFT CONFIG"""
REDSHIFT_CONF = {
    postgres._CONFIG_USERNAME_KEY: "",
    postgres._CONFIG_PASSWORD_KEY: "",
    postgres._CONFIG_DATABASE_KEY: "",
    postgres._CONFIG_HOST_KEY: "",
    postgres._CONFIG_PORT_KEY: "",
}

"""MARIADB CONFIG"""
MARIADB_CONF = {
    mysql._CONFIG_USERNAME_KEY: "",
    mysql._CONFIG_PASSWORD_KEY: "",
    mysql._CONFIG_DATABASE_KEY: "",
    mysql._CONFIG_HOST_KEY: "",
    mysql._CONFIG_PORT_KEY: "",
}

"""SQL SERVER CONFIG"""
SQL_SERVER_CONF = {
    sql_server._CONFIG_USERNAME_KEY: "",
    sql_server._CONFIG_PASSWORD_KEY: "",
    sql_server._CONFIG_DATABASE_KEY: "",
    sql_server._CONFIG_HOST_KEY: "",
    sql_server._CONFIG_PORT_KEY: "",
}

"""BIGQUERY CONFIG"""
BIGQUERY_CONF = {
    bigquery._CONFIG_PROJECT_ID_KEY: "",
    bigquery._CONFIG_SERVICE_ACCT_CREDENTIALS_KEY: "",
}

"""SQLITE CONFIG"""
SQLITE_CONF = {
    sqlite._CONFIG_DATABASE_PATH_KEY: "",
}



import pytest
from google.cloud import bigquery

from aqueduct_executor.operators.connectors.tabular import bigquery as bq
from aqueduct_executor.operators.connectors.tabular import dataframe

from aqueduct_executor.operators.connectors.tests import conf
from aqueduct_executor.operators.connectors.tests import utils

_DATASET = "testdata"
_TABLE = "testdata.bigquery"


@pytest.mark.skipif(conf.SKIP_BIGQUERY, reason="Skip BigQuery Flag Set")
class TestBigQuery:
    @classmethod
    def setup_class(cls):
        # Setup connector
        config = conf.BIGQUERY_CONF
        conn = bq.BigQueryConnector(config)
        cls.conn = conn

        # Setup test dataframe
        cls.test_df = utils.sample_df()

        # Create BigQuery client to perform setup and teardown
        cls.client = bigquery.Client(project=conn.project_id, credentials=conn.credentials)
        cls.client.create_dataset(_DATASET)

    @classmethod
    def teardown_class(cls):
        cls.client.delete_table(_TABLE)
        cls.client.delete_dataset(_DATASET)

    def test_authenticate(self):
        utils.authenticate_test(self.conn)

    @pytest.mark.dependency()
    def test_load(self):
        params = {dataframe.LOAD_PARAMS_TABLE_KEY: _TABLE}
        utils.load_test(self.conn, params, self.test_df)

    @pytest.mark.dependency(depends=["TestBigQuery::test_load"])
    def test_extract(self):
        params = {dataframe.EXTRACT_PARAMS_QUERY_KEY: "SELECT * FROM {};".format(_TABLE)}
        utils.extract_test(self.conn, params, expected_df=self.test_df)



import json
from typing import List

try:
    from typing import Literal
except ImportError:
    # Python 3.7 does not support typing.Literal
    from typing_extensions import Literal

from pydantic import BaseModel, Extra, parse_obj_as, validator
from aqueduct_executor.operators.utils import enums
from aqueduct_executor.operators.utils.storage import config


class FunctionSpec(BaseModel):
    name: str
    type: Literal[enums.JobType.FUNCTION]
    storage_config: config.StorageConfig
    metadata_path: str
    function_path: str
    function_extract_path: str
    entry_point_file: str
    entry_point_class: str
    entry_point_method: str
    custom_args: str
    input_content_paths: List[str]
    input_metadata_paths: List[str]
    output_content_paths: List[str]
    output_metadata_paths: List[str]
    input_artifact_types: List[enums.InputArtifactType]
    output_artifact_types: List[enums.OutputArtifactType]

    class Config:
        extra = Extra.forbid

    @validator("output_artifact_types")
    def check_metric_outputs(cls, output_artifact_types):
        if (
            len(output_artifact_types) > 1
            and enums.OutputArtifactType.FLOAT in output_artifact_types
        ):
            raise ValueError("A metric operator cannot have multiple outputs.")
        return output_artifact_types


def parse_spec(spec_json: str) -> FunctionSpec:
    """
    Parses a JSON string into a FunctionSpec.
    """
    data = json.loads(spec_json)

    return parse_obj_as(FunctionSpec, data)



import io
import json
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd

from aqueduct_executor.operators.utils.enums import InputArtifactType, OutputArtifactType
from aqueduct_executor.operators.utils.storage.storage import Storage

_DEFAULT_ENCODING = "utf8"


# Typing: all the possible artifact types to a function. Should be in sync with `InputArtifactType`.
InputArtifact = Union[pd.DataFrame, float, int]


def read_artifacts(
    storage: Storage,
    input_paths: List[str],
    input_metadata_paths: List[str],
    artifact_types: List[InputArtifactType],
) -> List[InputArtifact]:
    if len(input_paths) != len(artifact_types):
        raise Exception(
            "Found inconsistent number of input paths (%d) and artifact types (%d)"
            % (
                len(input_paths),
                len(artifact_types),
            )
        )

    inputs: List[InputArtifact] = []
    for (input_path, artifact_type) in zip(input_paths, artifact_types):
        if artifact_type == InputArtifactType.TABLE:
            inputs.append(_read_tabular_input(storage, input_path))
        elif artifact_type == InputArtifactType.FLOAT:
            # TODO(ENG-1119): A float artifact currently also represents integers.
            inputs.append(_read_numeric_input(storage, input_path))
        elif artifact_type == InputArtifactType.JSON:
            inputs.append(_read_json_input(storage, input_path))
        else:
            raise Exception("Unexpected input artifact type %s", artifact_type)
    return inputs


# TODO: Can also the input metadata here if we wanted to use it.
def _read_tabular_input(storage: Storage, path: str) -> pd.DataFrame:
    input_bytes = storage.get(path)
    return pd.read_json(io.BytesIO(input_bytes), orient="table")


def _read_numeric_input(storage: Storage, path: str) -> Union[float, int]:
    input_bytes = storage.get(path)

    # Check if it's an integer first, because casting a float to an int in this fashion
    # will throw a ValueError.
    try:
        return int(input_bytes)
    except ValueError:
        pass
    return float(input_bytes)


def _read_json_input(storage: Storage, path: str) -> Any:
    input_bytes = storage.get(path)
    return json.loads(input_bytes)


def write_artifacts(
    storage: Storage,
    output_paths: List[str],
    output_metadata_paths: List[str],
    contents: List[Any],
    artifact_types: List[OutputArtifactType],
) -> None:
    if (
        len(contents) != len(output_paths)
        or len(contents) != len(output_metadata_paths)
        or len(contents) != len(artifact_types)
    ):
        raise Exception(
            "Found inconsistent number of outputs (%d), artifact_types (%d), output paths (%d), and output metadata paths (%d)."
            % (
                len(contents),
                len(artifact_types),
                len(output_paths),
                len(output_metadata_paths),
            )
        )

    for (artifact_type, output_path, output_metadata_path, content) in zip(
        artifact_types, output_paths, output_metadata_paths, contents
    ):
        write_artifact(storage, output_path, output_metadata_path, content, artifact_type)


def write_artifact(
    storage: Storage,
    output_path: str,
    output_metadata_path: str,
    content: Any,
    artifact_type: OutputArtifactType,
) -> None:
    if artifact_type == OutputArtifactType.TABLE:
        if not isinstance(content, pd.DataFrame):
            raise Exception(
                "Expected output type to be Pandas Dataframe, but instead got %s"
                % type(content).__name__
            )
        _write_tabular_output(storage, output_path, output_metadata_path, content)
    elif artifact_type == OutputArtifactType.FLOAT:
        if not isinstance(content, float) and not isinstance(content, int):
            raise Exception(
                "Expected output type to be float or int, instead got %s" % type(content).__name__
            )
        _write_numeric_output(storage, output_path, output_metadata_path, content)
    elif artifact_type == OutputArtifactType.BOOL:
        if isinstance(content, bool) or isinstance(content, np.bool_):
            _write_bool_output(storage, output_path, output_metadata_path, bool(content))
        elif isinstance(content, pd.Series) and content.dtype == "bool":
            # We only write True if every boolean in the series is True.
            series = pd.Series(content)
            all_true = series.size - series.sum().item() == 0
            _write_bool_output(storage, output_path, output_metadata_path, all_true)
        else:
            raise Exception(
                "Expected output type to either a bool or a series of booleans, "
                "instead got %s" % type(content).__name__
            )
    elif artifact_type == OutputArtifactType.JSON:
        if not isinstance(content, str):
            raise Exception(
                "Expected output type to be string, instead got %s" % type(content).__name__
            )
        _write_json_output(storage, output_path, output_metadata_path, content)
    else:
        raise Exception("Unsupported output artifact type %s" % artifact_type)


def _write_tabular_output(
    storage: Storage,
    output_path: str,
    output_metadata_path: str,
    df: pd.DataFrame,
) -> None:
    output_str = df.to_json(orient="table", date_format="iso", index=False)

    # Create tabular output metadata
    schema = [{col: str(df[col].dtype)} for col in df]
    output_metadata_str = json.dumps(schema)

    storage.put(output_path, bytes(output_str, encoding=_DEFAULT_ENCODING))
    storage.put(output_metadata_path, bytes(output_metadata_str, encoding=_DEFAULT_ENCODING))


def _write_numeric_output(
    storage: Storage,
    output_path: str,
    output_metadata_path: str,
    val: Union[float, int],
) -> None:
    """Used for metrics."""
    storage.put(output_path, bytes(str(val), encoding=_DEFAULT_ENCODING))
    storage.put(output_metadata_path, bytes(json.dumps([]), encoding=_DEFAULT_ENCODING))


def _write_bool_output(
    storage: Storage,
    output_path: str,
    output_metadata_path: str,
    val: bool,
) -> None:
    """Used for checks."""
    storage.put(output_path, bytes(str(val), encoding=_DEFAULT_ENCODING))
    storage.put(output_metadata_path, bytes(json.dumps([]), encoding=_DEFAULT_ENCODING))


def _write_json_output(
    storage: Storage,
    output_path: str,
    output_metadata_path: str,
    val: str,
) -> None:
    """Used for parameters."""
    storage.put(output_path, bytes(val, encoding=_DEFAULT_ENCODING))
    storage.put(output_metadata_path, bytes(json.dumps([]), encoding=_DEFAULT_ENCODING))


def write_operator_metadata(
    storage: Storage,
    metadata_path: str,
    err: str,
    logs: Dict[str, str],
) -> None:
    """
    Writes operator execution metadata to storage.
    :param err: Any error message encountered during execution.
    :param logs: Any logs generated by this operator.
    """
    metadata: Dict[str, Any] = {"error": err, "logs": logs}
    storage.put(metadata_path, bytes(json.dumps(metadata), encoding=_DEFAULT_ENCODING))


def write_discover_results(storage: Storage, path: str, tables: List[str]):
    table_names_str = json.dumps(tables)

    storage.put(path, bytes(table_names_str, encoding=_DEFAULT_ENCODING))



from abc import ABC, abstractmethod
from typing import List

import pandas as pd

from aqueduct_executor.operators.connectors.tabular import extract, load


class TabularConnector(ABC):
    @abstractmethod
    def authenticate(self) -> None:
        """Authenticates connector configuration. Raises a ConnectionError if there is an error."""
        pass

    @abstractmethod
    def discover(self) -> List[str]:
        """Discover items in the connection.

        Returns:
            A list of items discovered.
        """
        pass

    @abstractmethod
    def extract(self, params: extract.Params) -> pd.DataFrame:
        """Extracts data from source into a DataFrame.

        Args:
            params: Extract parameters for the connector.

        Returns:
            A DataFrame that contains the data extracted by the connector.
        """
        pass

    @abstractmethod
    def load(self, params: load.Params, df: pd.DataFrame) -> None:
        """Loads DataFrame into destination.

        Args:
            params: Load parameters for the connector.
            df: DataFrame to load.
        """
        pass



from typing import Any

import boto3
from botocore.config import Config as BotoConfig

from typing import Any

from aqueduct_executor.operators.utils.storage.config import S3StorageConfig
from aqueduct_executor.operators.utils.storage.storage import Storage


class S3Storage(Storage):
    _client: Any  # boto3 s3 client
    _config: S3StorageConfig

    def __init__(self, config: S3StorageConfig):
        self._client = boto3.client("s3", config=BotoConfig(region_name=config.region))
        self._config = config

    def put(self, key: str, value: bytes) -> None:
        print(f"writing to s3: {key}")
        self._client.put_object(Bucket=self._config.bucket, Key=key, Body=value)

    def get(self, key: str) -> bytes:
        print(f"reading from s3: {key}")
        return self._client.get_object(Bucket=self._config.bucket, Key=key)["Body"].read()



from enum import Enum, EnumMeta
from typing import Any, Iterable, cast


class MetaEnum(EnumMeta):
    """Allows to very easily check if strings are present in the enum, without a helper.
    Eg.
        if "Postgres" in ServiceType:
            ...
    """

    def __contains__(cls, item: Any) -> Any:
        return item in [v.value for v in cast(Iterable[Enum], cls.__members__.values())]


class JobType(str, Enum, metaclass=MetaEnum):
    FUNCTION = "function"
    AUTHENTICATE = "authenticate"
    EXTRACT = "extract"
    LOAD = "load"
    DISCOVER = "discover"
    PARAM = "param"


class InputArtifactType(str, Enum, metaclass=MetaEnum):
    TABLE = "table"
    FLOAT = "float"
    JSON = "json"


class OutputArtifactType(str, Enum, metaclass=MetaEnum):
    TABLE = "table"
    FLOAT = "float"
    BOOL = "boolean"
    JSON = "json"



from sqlalchemy import create_engine, engine

from aqueduct_executor.operators.connectors.tabular import config, relational


class SqliteConnector(relational.RelationalConnector):
    def __init__(self, config: config.SqliteConfig):
        conn_engine = _create_engine(config)
        super().__init__(conn_engine)


def _create_engine(config: config.SqliteConfig) -> engine.Engine:
    # SQLite Dialect:
    # https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#dialect-sqlite-pysqlite-connect
    url = "sqlite:///{database}".format(
        database=config.database,
    )
    return create_engine(url)



from typing import Optional, Union


from pydantic import Field

from aqueduct_executor.operators.connectors.tabular import models


class BigQueryConfig(models.BaseConfig):
    project_id: str
    service_account_credentials: str


class MySqlConfig(models.BaseConfig):
    username: str
    password: str
    database: str
    host: str
    port: str


class PostgresConfig(models.BaseConfig):
    username: str
    password: str
    database: str
    host: str
    port: Optional[str] = "5432"


class S3Config(models.BaseConfig):
    access_key_id: str
    secret_access_key: str
    bucket: str


class SnowflakeConfig(models.BaseConfig):
    username: str
    password: str
    account_identifier: str
    database: str
    warehouse: str
    db_schema: Optional[str] = Field("public", alias="schema")  # schema is a Pydantic keyword


class SqlServerConfig(models.BaseConfig):
    username: str
    password: str
    database: str
    host: str
    port: str


class SqliteConfig(models.BaseConfig):
    database: str


Config = Union[
    BigQueryConfig,
    MySqlConfig,
    PostgresConfig,
    S3Config,
    SnowflakeConfig,
    SqlServerConfig,
    SqliteConfig,
]



import io
import os
from typing import List

import boto3
import pandas as pd

from aqueduct_executor.operators.connectors.tabular import common, config, connector, extract, load


class S3Connector(connector.TabularConnector):
    def __init__(self, config: config.S3Config):
        self.s3 = boto3.resource(
            "s3",
            aws_access_key_id=config.access_key_id,
            aws_secret_access_key=config.secret_access_key,
        )

        self.bucket = config.bucket

    def authenticate(self) -> None:
        pass

    def discover(self) -> List[str]:
        raise Exception("Discover is not supported for S3.")

    def extract(self, params: extract.S3Params) -> pd.DataFrame:
        response = self.s3.Object(self.bucket, params.filepath).get()
        data = response["Body"].read()
        buf = io.BytesIO(data)

        if params.format == common.S3FileFormat.CSV:
            return pd.read_csv(buf)
        elif params.format == common.S3FileFormat.JSON:
            return pd.read_json(buf)
        elif params.format == common.S3FileFormat.PARQUET:
            return pd.read_parquet(buf)

        raise Exception("Unknown S3 file format %s" % format)

    def load(self, params: load.S3Params, df: pd.DataFrame) -> None:
        buf = io.BytesIO()

        if params.format == common.S3FileFormat.CSV:
            df.to_csv(buf, index=False)
        elif params.format == common.S3FileFormat.JSON:
            # Index cannot be False for `to.json` for default orient
            # See: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_json.html
            df.to_json(buf)
        elif params.format == common.S3FileFormat.PARQUET:
            df.to_parquet(buf, index=False)
        else:
            raise Exception("Unknown S3 file format %s" % format)

        self.s3.Object(self.bucket, params.filepath).put(Body=buf.getvalue())



import argparse
import base64
import json
import sys
import traceback

from pydantic import parse_obj_as

from aqueduct_executor.operators.connectors.tabular import common, config, connector, spec
from aqueduct_executor.operators.utils import enums, utils
from aqueduct_executor.operators.utils.storage.parse import parse_storage
from aqueduct_executor.operators.utils.storage.storage import Storage


def run(spec: spec.Spec, storage: Storage):
    """
    Runs one of the following connector operations:
    - authenticate
    - extract
    - load
    - discover

    Arguments:
    - spec: The spec provided for this operator.
    - storage: An execution storage to use for reading or writing artifacts.
    """

    op = setup_connector(spec.connector_name, spec.connector_config)

    if spec.type == enums.JobType.AUTHENTICATE:
        run_authenticate(op)
    elif spec.type == enums.JobType.EXTRACT:
        run_extract(spec, op, storage)
    elif spec.type == enums.JobType.LOAD:
        run_load(spec, op, storage)
    elif spec.type == enums.JobType.DISCOVER:
        run_discover(spec, op, storage)
    else:
        raise Exception("Unknown job: %s" % spec.type)


def run_authenticate(op: connector.TabularConnector):
    op.authenticate()


def run_extract(spec: spec.ExtractSpec, op: connector.TabularConnector, storage: Storage):
    df = op.extract(spec.parameters)
    utils.write_artifacts(
        storage,
        [spec.output_content_path],
        [spec.output_metadata_path],
        [df],
        [utils.OutputArtifactType.TABLE],
    )


def run_load(spec: spec.LoadSpec, op: connector.TabularConnector, storage: Storage):
    inputs = utils.read_artifacts(
        storage,
        [spec.input_content_path],
        [spec.input_metadata_path],
        [utils.InputArtifactType.TABLE],
    )
    if len(inputs) != 1:
        raise Exception("Expected 1 input artifact, but got %d" % len(inputs))
    op.load(spec.parameters, inputs[0])


def run_discover(spec: spec.DiscoverSpec, op: connector.TabularConnector, storage: Storage):
    tables = op.discover()
    utils.write_discover_results(storage, spec.output_content_path, tables)


def setup_connector(
    connector_name: common.Name, connector_config: config.Config
) -> connector.TabularConnector:
    if connector_name == common.Name.AQUEDUCT_DEMO or connector_name == common.Name.POSTGRES:
        from aqueduct_executor.operators.connectors.tabular.postgres import (
            PostgresConnector as OpConnector,
        )
    elif connector_name == common.Name.SNOWFLAKE:
        from aqueduct_executor.operators.connectors.tabular.snowflake import (
            SnowflakeConnector as OpConnector,
        )
    elif connector_name == common.Name.BIG_QUERY:
        from aqueduct_executor.operators.connectors.tabular.bigquery import (
            BigQueryConnector as OpConnector,
        )
    elif connector_name == common.Name.REDSHIFT:
        from aqueduct_executor.operators.connectors.tabular.redshift import (
            RedshiftConnector as OpConnector,
        )
    elif connector_name == common.Name.SQL_SERVER:
        from aqueduct_executor.operators.connectors.tabular.sql_server import (
            SqlServerConnector as OpConnector,
        )
    elif connector_name == common.Name.MYSQL:
        from aqueduct_executor.operators.connectors.tabular.mysql import (
            MySqlConnector as OpConnector,
        )
    elif connector_name == common.Name.MARIA_DB:
        from aqueduct_executor.operators.connectors.tabular.maria_db import (
            MariaDbConnector as OpConnector,
        )
    elif connector_name == common.Name.AZURE_SQL:
        from aqueduct_executor.operators.connectors.tabular.azure_sql import (
            AzureSqlConnector as OpConnector,
        )
    elif connector_name == common.Name.S3:
        from aqueduct_executor.operators.connectors.tabular.s3 import S3Connector as OpConnector
    elif connector_name == common.Name.SQLITE:
        from aqueduct_executor.operators.connectors.tabular.sqlite import (
            SqliteConnector as OpConnector,
        )
    else:
        raise Exception("Unknown connector name: %s" % connector_name)

    return OpConnector(config=connector_config)


def _parse_spec(spec_json: str) -> spec.Spec:
    """
    Parses a JSON string into a spec.Spec.
    """
    data = json.loads(spec_json)

    print("Job Spec: \n{}".format(json.dumps(data, indent=4)))

    return parse_obj_as(spec.Spec, data)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = _parse_spec(spec_json)

    print("Started %s job: %s" % (spec.type, spec.name))

    storage = parse_storage(spec.storage_config)

    try:
        run(spec, storage)
        # Write operator execution metadata
        utils.write_operator_metadata(storage, spec.metadata_path, err="", logs={})
    except Exception as e:
        traceback.print_exc()
        err_msg = str(e)
        utils.write_operator_metadata(storage, spec.metadata_path, err=err_msg, logs={})
        sys.exit(1)



import argparse
import base64
import traceback
import sys

from aqueduct_executor.operators.utils import enums, utils
from aqueduct_executor.operators.param_executor import spec
from aqueduct_executor.operators.utils.storage.parse import parse_storage


def run(spec: spec.ParamSpec) -> None:
    """
    Executes a parameter operator by storing the parameter value in the output content path.
    """
    storage = parse_storage(spec.storage_config)
    try:
        utils.write_artifact(
            storage,
            spec.output_content_path,
            spec.output_metadata_path,
            spec.val,
            enums.OutputArtifactType.JSON,
        )
        utils.write_operator_metadata(storage, spec.metadata_path, "", {})
    except Exception as e:
        utils.write_operator_metadata(storage, spec.metadata_path, str(e), {})
        print("Exception Raised: ", e)
        traceback.print_tb(e.__traceback__)
        sys.exit(1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = spec.parse_spec(spec_json)

    print("Job Spec: \n{}".format(spec.json()))
    run(spec)



import argparse
import base64
import importlib
import io
import json
import os
import sys
import traceback
from contextlib import redirect_stderr, redirect_stdout
from typing import Any, Callable, List, Tuple, Dict

from aqueduct_executor.operators.function_executor import spec
from aqueduct_executor.operators.function_executor.utils import OP_DIR
from aqueduct_executor.operators.utils import utils
from aqueduct_executor.operators.utils.storage.storage import Storage
from aqueduct_executor.operators.utils.storage.parse import parse_storage

from pandas import DataFrame


def _user_fn_traceback(offset: int = 0) -> str:
    """
    Captures the stack traceback and returns it as a string. If offset is positive,
    it will extract the traceback starting at OFFSET frames from the top (e.g. most recent frame).
    An offset of 1 means the most recent frame will be excluded.
    """
    file = io.StringIO()

    tb_type, tb_val, tb = sys.exc_info()
    while offset > 0:
        if tb is None or tb.tb_next is None:
            break
        tb = tb.tb_next
        offset -= 1

    traceback.print_exception(tb_type, tb_val, tb, file=file)

    file.seek(0)
    return file.read()


def _fetch_logs(stdout_log, stderr_log):
    """
    If there is any output, set as the values for protected keys STDOUT_KEY and STDERR_KEY.
    """
    stdout_log.seek(0)
    stderr_log.seek(0)

    logs = {}

    stdout_contents = stdout_log.read()
    if len(stdout_contents) > 0:
        logs["stdout"] = stdout_contents

    stderr_contents = stderr_log.read()
    if len(stderr_contents) > 0:
        logs["stderr"] = stderr_contents
    return logs


def _get_py_import_path(spec: spec.FunctionSpec) -> str:
    """
    Generates the import path based on fixed function dir and
    FUNCTION_ENTRY_POINT_FILE env var.

    It removes .py (if any) from the entry point and replaces all
    '/' with '.'

    For example, entry point 'model/churn.py'  will finally become
    'app.function.model.churn', where we can import from.
    """
    file_path = spec.entry_point_file
    if file_path.endswith(".py"):
        file_path = file_path[:-3]

    if file_path.startswith("/"):
        file_path = file_path[1:]
    return ".".join([OP_DIR, file_path.replace("/", ".")])


def _import_invoke_method(spec: spec.FunctionSpec) -> Callable[..., DataFrame]:
    fn_path = spec.function_extract_path
    os.chdir(os.path.join(fn_path, OP_DIR))
    sys.path.append(fn_path)
    import_path = _get_py_import_path(spec)
    class_name = spec.entry_point_class
    method_name = spec.entry_point_method
    custom_args_str = spec.custom_args
    # Invoke the function and parse out the result object.
    module = importlib.import_module(import_path)
    if not class_name:
        return getattr(module, method_name)

    fn_class = getattr(module, class_name)
    function = fn_class()
    # Set the custom arguments if provided
    if custom_args_str:
        custom_args = json.loads(custom_args_str)
        function.set_args(custom_args)

    return getattr(function, method_name)


def _execute_function(
    spec: spec.FunctionSpec,
    inputs: List[utils.InputArtifact],
) -> Tuple[Any, Dict[str, str], str]:
    """
    Invokes the given function on the input data. Does not raise an exception on any
    user function errors. Instead, returns the error message as a string.

    :param inputs: the input data to feed into the user's function.
    """
    stdout_log = io.StringIO()
    stderr_log = io.StringIO()

    try:
        invoke = _import_invoke_method(spec)
        print("Invoking the function...")
        try:
            with redirect_stdout(stdout_log), redirect_stderr(stderr_log):
                result = invoke(*inputs)  # Unpack DataFrames argument list
        except Exception:
            # Include the stack trace within the user's code.
            return (
                None,
                _fetch_logs(stdout_log, stderr_log),
                _user_fn_traceback(offset=1),
            )
    except Exception as e:
        # For user-induced errors that we can't extract stack traces for,
        # just return the exception message.
        return (
            None,
            _fetch_logs(stdout_log, stderr_log),
            str(type(e).__name__) + ": " + str(e),
        )
    finally:
        sys.path.pop(0)

    return result, _fetch_logs(stdout_log, stderr_log), ""


def _write_outputs(
    spec: spec.FunctionSpec,
    storage: Storage,
    res: Any,
    logs: Dict[str, str],
):
    # Force all results to be of type `list`, so we can always loop over them.
    results = res
    if not isinstance(res, list):
        results = [res]

    utils.write_artifacts(
        storage,
        spec.output_content_paths,
        spec.output_metadata_paths,
        results,
        spec.output_artifact_types,
    )
    utils.write_operator_metadata(storage, spec.metadata_path, "", logs)


def run(spec: spec.FunctionSpec) -> None:
    """
    Executes a function operator.
    """
    storage = parse_storage(spec.storage_config)
    logs = {}
    try:
        # Read the input data from intermediate storage.
        inputs = utils.read_artifacts(
            storage, spec.input_content_paths, spec.input_metadata_paths, spec.input_artifact_types
        )

        print("Invoking the function...")
        results, logs, err_msg = _execute_function(spec, inputs)
        if len(err_msg) > 0:
            raise Exception(err_msg)

        print("Function invoked successfully!")

        # Force all results to be of type `list`, so we can always loop over them.
        if not isinstance(results, list):
            results = [results]

        utils.write_artifacts(
            storage,
            spec.output_content_paths,
            spec.output_metadata_paths,
            results,
            spec.output_artifact_types,
        )
        utils.write_operator_metadata(storage, spec.metadata_path, "", logs)

    except Exception as e:
        utils.write_operator_metadata(storage, spec.metadata_path, str(e), logs)
        print("Exception Raised: ", e)
        traceback.print_tb(e.__traceback__)
        sys.exit(1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = spec.parse_spec(spec_json)

    print("Started %s job: %s" % (spec.type, spec.name))

    run(spec)



from typing import List

import pandas as pd
from sqlalchemy import engine, inspect
from sqlalchemy.exc import SQLAlchemyError

from aqueduct_executor.operators.connectors.tabular import connector, extract, load


class RelationalConnector(connector.TabularConnector):
    def __init__(self, conn_engine: engine.Engine):
        self.engine = conn_engine

    def __del__(self):
        self.engine.dispose()

    def authenticate(self) -> None:
        try:
            self.engine.connect()
        except SQLAlchemyError as e:
            raise ConnectionError("Unable to connect.") from e

    def discover(self) -> List[str]:
        return inspect(self.engine).get_table_names()

    def extract(self, params: extract.RelationalParams) -> pd.DataFrame:
        df = pd.read_sql(params.query, con=self.engine)
        return df

    def load(self, params: load.RelationalParams, df: pd.DataFrame) -> None:
        # NOTE (saurav): df._to_sql has known performance issues. Using `method="multi"` helps incrementally,
        # since pandas will pass multiple rows in a single INSERT. If this still remains an issue, we can pass in a
        # callable function for `method` that does bulk loading.
        # See: https://pandas.pydata.org/docs/user_guide/io.html#io-sql-method
        df.to_sql(
            params.table,
            con=self.engine,
            if_exists=params.update_mode.value,
            index=False,
            method="multi",
        )



from typing import Any, Optional, Union

from aqueduct_executor.operators.connectors.tabular import common, models


class RelationalParams(models.BaseParams):
    query: str
    # TODO: Consider not including github as part of relational params when it is JSON marshalled
    github_metadata: Optional[Any]


class S3Params(models.BaseParams):
    filepath: str
    format: common.S3FileFormat


Params = Union[RelationalParams, S3Params]



import argparse
import base64
import os

from aqueduct_executor.operators.function_executor.utils import OP_DIR
from . import spec


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-s", "--spec", required=True)
    args = parser.parse_args()

    spec_json = base64.b64decode(args.spec)
    spec = spec.parse_spec(spec_json)
    # The output of the print statement to stdout is captured by the calling bash script into a variable,
    # so we should not include any other print statements in this Python script.
    print(os.path.join(spec.function_extract_path, OP_DIR))



import json
from typing import List

import pandas as pd
import pandas_gbq
from google.oauth2 import service_account

from aqueduct_executor.operators.connectors.tabular import config, connector, extract, load


class BigQueryConnector(connector.TabularConnector):
    def __init__(self, config: config.BigQueryConfig):
        self.project_id = config.project_id

        credentials_info = json.loads(config.service_account_credentials)
        self.credentials = service_account.Credentials.from_service_account_info(credentials_info)

    def authenticate(self) -> None:
        # pandas_gbq does not have an explicit authenticate method, so we execute a test query
        pandas_gbq.read_gbq("SELECT 1;", project_id=self.project_id, credentials=self.credentials)

    def discover(self) -> List[str]:
        # TBD eng-708-investigate-how-to-list-tables-for-bigquery
        return []

    def extract(self, params: extract.RelationalParams) -> pd.DataFrame:
        df = pandas_gbq.read_gbq(
            params.query, project_id=self.project_id, credentials=self.credentials
        )
        return df

    def load(self, params: load.RelationalParams, df: pd.DataFrame) -> None:
        pandas_gbq.to_gbq(
            df,
            params.table,
            project_id=self.project_id,
            if_exists=params.update_mode.value,
            credentials=self.credentials,
        )



from typing import Union

try:
    from typing import Literal
except ImportError:
    # Python 3.7 does not support typing.Literal
    from typing_extensions import Literal

from pydantic import validator

from aqueduct_executor.operators.connectors.tabular import common, config, extract, load, models
from aqueduct_executor.operators.utils import enums
from aqueduct_executor.operators.utils.storage import config as sconfig


def unwrap_connector_config(cls, connector_config, values):
    """
    TODO ENG-937: Remove this validator once connector config serialization is fixed.

    Unwraps the connector config before it can be parsed into a
    config.Config object. This is necessary because of how connector_config
    is serialized in Golang.

    For non-OAuth configs, it has the following structure:
    "connector_config": {
        "conf": {
            "username": "username",
            "password": "password",
        }
    }

    For OAuth configs, it has the following structure:
    "connector_config": {
        "token": {
            "access_token": "123456",
            "refresh_token": "123",
        },
        "oauth2_conf": {...},
        "public_conf": {...},
    }
    """

    if "connector_name" not in values:
        raise ValueError("Unknown connector name.")

    connector_name = values["connector_name"]

    if not isinstance(connector_config, dict):
        raise ValueError("connector_config is not a dictionary.")

    # This is a static config
    return connector_config["conf"]


class AuthenticateSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.AUTHENTICATE]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class ExtractSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.EXTRACT]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config
    parameters: extract.Params
    output_content_path: str
    output_metadata_path: str

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class LoadSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.LOAD]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config
    parameters: load.Params
    input_content_path: str
    input_metadata_path: str

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


class DiscoverSpec(models.BaseSpec):
    name: str
    type: Literal[enums.JobType.DISCOVER]
    storage_config: sconfig.StorageConfig
    metadata_path: str
    connector_name: common.Name
    connector_config: config.Config
    output_content_path: str

    # validators
    _unwrap_connector_config = validator("connector_config", allow_reuse=True, pre=True)(
        unwrap_connector_config
    )


Spec = Union[AuthenticateSpec, ExtractSpec, LoadSpec, DiscoverSpec]



from typing import Optional, Union

from pydantic import validator

from aqueduct_executor.operators.connectors.tabular import common, models


class RelationalParams(models.BaseParams):
    table: str
    update_mode: Optional[common.UpdateMode] = common.UpdateMode.REPLACE

    class Config:
        validate_assignment = True

    @validator("update_mode")
    def set_update_mode(cls, update_mode):
        if update_mode == "":
            return common.UpdateMode.REPLACE
        return update_mode


class S3Params(models.BaseParams):
    filepath: str
    format: common.S3FileFormat


Params = Union[RelationalParams, S3Params]
